class Matrix:
    def __init__(self, data: list[list]):

        self.validate(data)
        self.data = data

    @staticmethod
    def validate(data):
        try:
            assert isinstance(data, list)
            assert all([isinstance(row, list) for row in data])
            row_lens = [len(row) for row in data]
            assert all([row_lens[0] == row_len for row_len in row_lens])
            assert all([isinstance(el, (float, int))
                       for row in data for el in row])
        except AssertionError:
            raise ValueError

    def num_rows(self):
        return len(self.data)

    def num_cols(self):
        return len(self.data[0])

    def get_value(self, row_number, col_number):
        return self.data[row_number][col_number]

    def __add__(self, other):
        self.assert_shape_equal(other)

        new_data = []
        for row_1, row_2 in zip(self.data, other.data):
            new_data.append([i+j for i, j in zip(row_1, row_2)])
        return Matrix(new_data)

    def __eq__(self, other: object) -> bool:
        return self.data == other.data

    def assert_shape_equal(self, other):
        try:
            assert self.num_cols() == other.num_cols()
            assert self.num_rows() == other.num_rows()
        except AssertionError:
            raise ValueError

    def hadamard_product(self, other):
        self.assert_shape_equal(other)
        new_data = []
        for row_1, row_2 in zip(self.data, other.data):
            new_data.append([i*j for i, j in zip(row_1, row_2)])
        return Matrix(new_data)

    def transpose(self):
        transposed_data = [[] for _ in range(self.num_cols())]
        for row in self.data:
            for i, value in enumerate(row):
                transposed_data[i].append(value)
        return Matrix(transposed_data)

    def assert_square(self):
        try:
            assert self.num_cols() == self.num_rows()
        except:
            raise ValueError

    @staticmethod
    def dot_product(vec_1, vec_2):
        return sum(i*j for i, j in zip(vec_1, vec_2))

    def __mul__(self, other):
        try:
            assert self.num_cols() == other.num_rows()
        except:
            raise ValueError
        result = [[0 for _ in range(self.num_cols())]
                  for _ in range(self.num_rows())]
        first = self.data
        second = other.transpose().data
        for i in range(self.num_rows()):
            for j in range(other.num_cols()):
                result[i][j] = self.dot_product(first[i], second[j])
        return Matrix(result)

    def LU_decompose(self):
        self.assert_square()
        U = [[0 for _ in range(self.num_cols())]
             for _ in range(self.num_rows())]
        L = [[0 if i != j else 1 for i in range(
            self.num_cols())] for j in range(self.num_rows())]

        shape = self.num_cols()
        for i in range(shape):
            for j in range(shape):
                if j >= i:
                    if i == 0:
                        U[i][j] = self.data[i][j]
                    else:
                        U[i][j] = self.data[i][j] - \
                            sum([L[i][k]*U[k][j] for k in range(i)])
            for j in range(shape):
                if j >= i:
                    if i == 0:
                        L[j][i] = self.data[j][i]/U[i][i]
                    else:
                        L[j][i] = (self.data[j][i] - sum(L[j][k]*U[k][i]
                                   for k in range(i)))/U[i][i]
        return L, U
    
    def determinant(self):
        L, U = self.LU_decompose()
        L_det = sum(l[i] for i, l in enumerate(L))
        U_det = sum(u[i] for i, u in enumerate(U))
        return L_det*U_det
